#  tests for pymks-0.4.1-py37h03978a9_1 (this is a generated file);
print('===== testing package: pymks-0.4.1-py37h03978a9_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import pymks

pymks.test()
#  --- run_test.py (end) ---

print('===== pymks-0.4.1-py37h03978a9_1 OK =====');
print("import: 'pymks'")
import pymks

